import copy
from dataclasses import dataclass
from enum import IntEnum
from typing import List, Optional

import numpy as np
import numpy.typing as npt
from nuplan.common.maps.abstract_map_objects import (
    GraphEdgeMapObject,
    Lane,
    LaneConnector,
    LaneGraphEdgeMapObject,
    PolylineMapObject,
)
from nuplan.common.actor_state.state_representation import Point2D
from nuplan.common.maps.abstract_map import AbstractMap
from nuplan.common.maps.maps_datatypes import SemanticMapLayer
from nuplan.common.actor_state.state_representation import StateSE2
from nuplan.common.actor_state.tracked_objects_types import AGENT_TYPES
from nuplan.common.actor_state.vehicle_parameters import VehicleParameters, get_pacifica_parameters
from nuplan.common.maps.maps_datatypes import SemanticMapLayer
from nuplan.planning.metrics.utils.collision_utils import CollisionType
from nuplan.planning.simulation.observation.idm.utils import (
    is_agent_ahead,
    is_agent_behind,
)
from nuplan.planning.simulation.trajectory.trajectory_sampling import TrajectorySampling
from shapely import Point, creation
from shapely.geometry import Point
from navsim.agents.expansion.submetrics.metric_lk import calc_lk
from navsim.agents.expansion.submetrics.metric_tl import calc_tl
from navsim.agents.expansion.submetrics.metric_consistency import calc_consistency
from navsim.agents.expansion.submetrics.metric_direction import calc_driving_direction_compliance
from navsim.planning.metric_caching.metric_cache import MetricCache
from navsim.planning.simulation.planner.pdm_planner.observation.pdm_observation import (
    PDMObservation,
)
from navsim.planning.simulation.planner.pdm_planner.observation.pdm_occupancy_map import (
    PDMDrivableMap, PDMCrosswalkIntersectionMap,
)
from navsim.planning.simulation.planner.pdm_planner.scoring.pdm_comfort_metrics import (
    ego_is_comfortable,
)
from navsim.planning.simulation.planner.pdm_planner.scoring.pdm_scorer_utils import (
    get_collision_type,
)
from navsim.planning.simulation.planner.pdm_planner.utils.pdm_array_representation import (
    coords_array_to_polygon_array,
    state_array_to_coords_array,
)
from navsim.planning.simulation.planner.pdm_planner.utils.pdm_enums import (
    BBCoordsIndex,
    EgoAreaIndex,
    StateIndex,
)
from navsim.planning.simulation.planner.pdm_planner.utils.pdm_path import PDMPath


class MultiMetricIndex(IntEnum):
    """Index mapping multiplicative metrics (used in PDMScorer)."""

    NO_COLLISION = 0
    DRIVABLE_AREA = 1
    DRIVING_DIRECTION = 2
    TRAFFIC_LIGHTS = 3


class WeightedMetricIndex(IntEnum):
    """Index mapping weighted metrics (used in PDMScorer)."""

    PROGRESS = 0
    TTC = 1
    COMFORTABLE = 2
    LANE_KEEPING = 3
    CONSISTENCY = 4


@dataclass
class PDMScorerConfigExpanded:
    # weighted metric weights
    progress_weight: float = 5.0
    ttc_weight: float = 5.0
    comfortable_weight: float = 2.0
    lk_weight: float = 5.0

    # thresholds
    driving_direction_horizon: float = 1.0  # [s] (driving direction)
    driving_direction_compliance_threshold: float = 0.5  # [m] (driving direction)
    driving_direction_violation_threshold: float = 1.5  # [m] (driving direction)
    stopped_speed_threshold: float = 5e-03  # [m/s] (ttc)
    progress_distance_threshold: float = 0.1  # [m] (progress)

    @property
    def weighted_metrics_array(self) -> npt.NDArray[np.float64]:
        weighted_metrics = np.zeros(len(WeightedMetricIndex), dtype=np.float64)
        weighted_metrics[WeightedMetricIndex.PROGRESS] = self.progress_weight
        weighted_metrics[WeightedMetricIndex.TTC] = self.ttc_weight
        weighted_metrics[WeightedMetricIndex.COMFORTABLE] = self.comfortable_weight
        # weighted_metrics[WeightedMetricIndex.LANE_KEEPING] = self.lk_weight
        return weighted_metrics

    @property
    def weighted_metrics_array_expand(self) -> npt.NDArray[np.float64]:
        weighted_metrics = np.zeros(len(WeightedMetricIndex), dtype=np.float64)
        weighted_metrics[WeightedMetricIndex.PROGRESS] = self.progress_weight
        weighted_metrics[WeightedMetricIndex.TTC] = self.ttc_weight
        weighted_metrics[WeightedMetricIndex.COMFORTABLE] = self.comfortable_weight
        weighted_metrics[WeightedMetricIndex.LANE_KEEPING] = self.lk_weight
        return weighted_metrics


class PDMScorerExpanded:
    """Class to score proposals in PDM pipeline. Re-implements nuPlan's closed-loop metrics."""

    def __init__(
            self,
            proposal_sampling: TrajectorySampling,
            config: PDMScorerConfigExpanded = PDMScorerConfigExpanded(),
            vehicle_parameters: VehicleParameters = get_pacifica_parameters(),
    ):
        """
        Constructor of PDMScorer
        :param proposal_sampling: Sampling parameters for proposals
        """
        self.proposal_sampling = proposal_sampling
        self._config = config
        self._vehicle_parameters = vehicle_parameters

        # lazy loaded
        self._observation: Optional[PDMObservation] = None
        self._centerline: Optional[PDMPath] = None
        self._route_lane_ids: Optional[List[str]] = None
        self._drivable_area_map: Optional[PDMDrivableMap] = None

        self._num_proposals: Optional[int] = None
        self._states: Optional[npt.NDArray[np.float64]] = None
        self._ego_coords: Optional[npt.NDArray[np.float64]] = None
        self._ego_polygons: Optional[npt.NDArray[np.object_]] = None

        self._ego_areas: Optional[npt.NDArray[np.bool_]] = None

        self._multi_metrics: Optional[npt.NDArray[np.float64]] = None
        self._weighted_metrics: Optional[npt.NDArray[np.float64]] = None
        self._progress_raw: Optional[npt.NDArray[np.float64]] = None

        self._collision_time_idcs: Optional[npt.NDArray[np.float64]] = None
        self._ttc_time_idcs: Optional[npt.NDArray[np.float64]] = None

        self.map_api = None
    def time_to_at_fault_collision(self, proposal_idx: int) -> float:
        """
        Returns time to at-fault collision for given proposal
        :param proposal_idx: index for proposal
        :return: time to infraction
        """
        return self._collision_time_idcs[proposal_idx] * self.proposal_sampling.interval_length

    def time_to_ttc_infraction(self, proposal_idx: int) -> float:
        """
        Returns time to ttc infraction for given proposal
        :param proposal_idx: index for proposal
        :return: time to infraction
        """
        return self._ttc_time_idcs[proposal_idx] * self.proposal_sampling.interval_length

    def score_proposals(
            self,
            states: npt.NDArray[np.float64],
            pre_states,
            observation: PDMObservation,
            centerline: PDMPath,
            route_lane_ids: List[str],
            drivable_area_map: PDMDrivableMap,
            metric_cache: MetricCache,
            expansion_only: bool
    ) -> npt.NDArray[np.float64]:
        """
        Scores proposal similar to nuPlan's closed-loop metrics
        :param states: array representation of simulated proposals
        :param observation: PDM's observation class
        :param centerline: path of the centerline
        :param route_lane_ids: list containing on-route lane ids
        :param drivable_area_map: Occupancy map of drivable are polygons
        :return: array containing score of each proposal
        """

        # initialize & lazy load class values
        self._reset(
            states,
            pre_states,
            observation,
            centerline,
            route_lane_ids,
            drivable_area_map,
        )
        self.metric_cache = metric_cache
        # fill value ego-area array (used in multiple metrics)
        self._calculate_ego_area()

        # calc expansion
        if expansion_only:
            # self._calculate_driving_direction_compliance()
            self.calc_lk()
            # self.calc_tl()
            self.navigation_mAP = self.calc_map()
            return None

        # calc everything
        # 1. multiplicative metrics
        self._calculate_no_at_fault_collision()
        self._calculate_drivable_area_compliance()
        self._calculate_driving_direction_compliance()

        # 2. weighted metrics
        self._calculate_progress()
        self._calculate_ttc()
        self._calculate_is_comfortable()

        # 3. expanded metrics
        self.calc_lk()
        self.calc_tl()
        self.calc_consistency()
        self.navigation_mAP = self.calc_map()

        # todo new scores arent' aggregated into final pdms yet
        return self._aggregate_scores(), self._aggregate_scores_expand()

    def calc_lk(self):
        self._weighted_metrics[WeightedMetricIndex.LANE_KEEPING] = calc_lk(
            self._ego_coords,
            self._centerline,
            BBCoordsIndex,
            self._num_proposals,
            self._multi_metrics,
        )

    def calc_tl(self):
        # self._multi_metrics[MultiMetricIndex.TRAFFIC_LIGHTS] = 1.0
        self._multi_metrics[MultiMetricIndex.TRAFFIC_LIGHTS] = calc_tl(
            self._states,
            self._num_proposals,
            self._drivable_area_map,
            self.metric_cache,
            self._centerline,
            self._route_lane_ids,
            self._config,
        )

    def calc_consistency(self):
        # self._weighted_metrics[WeightedMetricIndex.CONSISTENCY] = calc_consistency(
        #     self._ego_coords,
        #     self._pre_ego_coords,
        #     BBCoordsIndex,
        #     self._num_proposals,
        #     self._multi_metrics,
        # )
        self._weighted_metrics[WeightedMetricIndex.CONSISTENCY] = 1.0

    def calc_map(self):
        return 1.0

    def _aggregate_scores(self) -> npt.NDArray[np.float64]:
        """
        Aggregates metrics with multiplicative and weighted average.
        :return: array containing score of each proposal
        """
        # accumulate multiplicative metrics
        # multiplicate_metric_scores = self._multi_metrics.prod(axis=0)
        # todo ignore tl metric
        multiplicate_metric_scores = self._multi_metrics[:3].prod(axis=0)

        # normalize and fill progress values
        raw_progress = self._progress_raw * multiplicate_metric_scores
        N = raw_progress.shape[0]
        pdm_progress = np.repeat(raw_progress[0], N)[..., None]
        combined_progress = np.concatenate([raw_progress[..., None], pdm_progress], axis=1)
        max_raw_progress = np.max(
            combined_progress,
            axis=1
        )
        # three cases:
        # 1. bigger than t ---------- normalize
        # 2. smaller than t & score!=0 -------- 1
        # 3. smaller than t & score==0 -------- 0
        bigger_than_t_mask = max_raw_progress > self._config.progress_distance_threshold  # (4096,1)
        smaller_than_t_mask = np.logical_not(bigger_than_t_mask)
        bad_mask = multiplicate_metric_scores == 0.0
        smaller_and_bad = np.logical_and(bad_mask, smaller_than_t_mask)

        normalized_progress = np.ones_like(raw_progress)
        normalized_progress[smaller_and_bad] = 0.0
        normalized_progress[bigger_than_t_mask] = raw_progress[bigger_than_t_mask] / max_raw_progress[
            bigger_than_t_mask]

        # max_raw_progress = np.max(raw_progress)
        # if max_raw_progress > self._config.progress_distance_threshold:
        #     normalized_progress = raw_progress / max_raw_progress
        # else:
        #     normalized_progress = np.ones(len(raw_progress), dtype=np.float64)
        #     normalized_progress[multiplicate_metric_scores == 0.0] = 0.0
        self._weighted_metrics[WeightedMetricIndex.PROGRESS] = normalized_progress

        # accumulate weighted metrics
        weighted_metrics_array = self._config.weighted_metrics_array
        weighted_metric_scores = (self._weighted_metrics * weighted_metrics_array[..., None]).sum(
            axis=0
        )
        weighted_metric_scores /= weighted_metrics_array.sum()

        # calculate final scores
        final_scores = self._multi_metrics.prod(axis=0) * weighted_metric_scores

        return final_scores

    def _aggregate_scores_expand(self) -> npt.NDArray[np.float64]:
        """
        Aggregates metrics with multiplicative and weighted average.
        :return: array containing score of each proposal
        """
        # accumulate multiplicative metrics
        # multiplicate_metric_scores = self._multi_metrics.prod(axis=0)
        # todo ignore tl metric
        multiplicate_metric_scores = self._multi_metrics[:4].prod(axis=0)

        # normalize and fill progress values
        raw_progress = self._progress_raw * multiplicate_metric_scores
        N = raw_progress.shape[0]
        pdm_progress = np.repeat(raw_progress[0], N)[..., None]
        combined_progress = np.concatenate([raw_progress[..., None], pdm_progress], axis=1)
        max_raw_progress = np.max(
            combined_progress,
            axis=1
        )
        # three cases:
        # 1. bigger than t ---------- normalize
        # 2. smaller than t & score!=0 -------- 1
        # 3. smaller than t & score==0 -------- 0
        bigger_than_t_mask = max_raw_progress > self._config.progress_distance_threshold  # (4096,1)
        smaller_than_t_mask = np.logical_not(bigger_than_t_mask)
        bad_mask = multiplicate_metric_scores == 0.0
        smaller_and_bad = np.logical_and(bad_mask, smaller_than_t_mask)

        normalized_progress = np.ones_like(raw_progress)
        normalized_progress[smaller_and_bad] = 0.0
        normalized_progress[bigger_than_t_mask] = raw_progress[bigger_than_t_mask] / max_raw_progress[
            bigger_than_t_mask]

        # max_raw_progress = np.max(raw_progress)
        # if max_raw_progress > self._config.progress_distance_threshold:
        #     normalized_progress = raw_progress / max_raw_progress
        # else:
        #     normalized_progress = np.ones(len(raw_progress), dtype=np.float64)
        #     normalized_progress[multiplicate_metric_scores == 0.0] = 0.0
        self._weighted_metrics[WeightedMetricIndex.PROGRESS] = normalized_progress

        # accumulate weighted metrics
        weighted_metrics_array = self._config.weighted_metrics_array_expand
        weighted_metric_scores = (self._weighted_metrics * weighted_metrics_array[..., None]).sum(
            axis=0
        )
        weighted_metric_scores /= weighted_metrics_array.sum()

        # calculate final scores
        final_scores = self._multi_metrics.prod(axis=0) * weighted_metric_scores

        return final_scores

    def _reset(
            self,
            states: npt.NDArray[np.float64],
            pre_states,
            observation: PDMObservation,
            centerline: PDMPath,
            route_lane_ids: List[str],
            drivable_area_map: PDMDrivableMap,
    ) -> None:
        """
        Resets metric values and lazy loads input classes.
        :param states: array representation of simulated proposals
        :param observation: PDM's observation class
        :param centerline: path of the centerline
        :param route_lane_ids: list containing on-route lane ids
        :param drivable_area_map: Occupancy map of drivable are polygons
        """
        assert states.ndim == 3
        assert states.shape[1] == self.proposal_sampling.num_poses + 1
        assert states.shape[2] == StateIndex.size()

        self._observation = observation
        self._centerline = centerline
        self._route_lane_ids = route_lane_ids
        self._drivable_area_map = drivable_area_map
        self._num_proposals = states.shape[0]

        # save ego state values
        self._states = states
        self._pre_states = pre_states

        # calculate coordinates of ego corners and center
        self._ego_coords = state_array_to_coords_array(states, self._vehicle_parameters)
        if pre_states is not None:
            self._pre_ego_coords = state_array_to_coords_array(pre_states, self._vehicle_parameters)
        else:
            self._pre_ego_coords = None

        # initialize all ego polygons from corners
        self._ego_polygons = coords_array_to_polygon_array(self._ego_coords)

        # zero initialize all remaining arrays.
        self._ego_areas = np.zeros(
            (
                self._num_proposals,
                self.proposal_sampling.num_poses + 1,
                len(EgoAreaIndex),
            ),
            dtype=np.bool_,
        )
        self._multi_metrics = np.zeros(
            (len(MultiMetricIndex), self._num_proposals), dtype=np.float64
        )
        self._weighted_metrics = np.zeros(
            (len(WeightedMetricIndex), self._num_proposals), dtype=np.float64
        )
        self._progress_raw = np.zeros(self._num_proposals, dtype=np.float64)

        # initialize infraction arrays with infinity (meaning no infraction occurs)
        self._collision_time_idcs = np.zeros(self._num_proposals, dtype=np.float64)
        self._ttc_time_idcs = np.zeros(self._num_proposals, dtype=np.float64)
        self._collision_time_idcs.fill(np.inf)
        self._ttc_time_idcs.fill(np.inf)

    def _calculate_ego_area(self) -> None:
        """
        Determines the area of proposals over time.
        Areas are (1) in multiple lanes, (2) non-drivable area, or (3) oncoming traffic
        """

        n_proposals, n_horizon, n_points, _ = self._ego_coords.shape

        in_polygons = self._drivable_area_map.points_in_polygons(self._ego_coords)
        in_polygons = in_polygons.transpose(
            1, 2, 0, 3
        )  # shape: n_proposals, n_horizon, n_polygons, n_points

        drivable_area_idcs = self._drivable_area_map.get_indices_of_map_type(
            [
                SemanticMapLayer.ROADBLOCK,
                SemanticMapLayer.INTERSECTION,
                SemanticMapLayer.DRIVABLE_AREA,
                SemanticMapLayer.CARPARK_AREA,
            ]
        )

        drivable_lane_idcs = self._drivable_area_map.get_indices_of_map_type(
            [SemanticMapLayer.LANE, SemanticMapLayer.LANE_CONNECTOR]
        )

        drivable_on_route_idcs: List[int] = [
            idx
            for idx in drivable_lane_idcs
            if self._drivable_area_map.tokens[idx] in self._route_lane_ids
        ]  # index mask for on-route lanes

        corners_in_polygon = in_polygons[..., :-1]  # ignore center coordinate
        center_in_polygon = in_polygons[..., -1]  # only center

        # in_multiple_lanes: if
        # - more than one drivable polygon contains at least one corner
        # - no polygon contains all corners
        batch_multiple_lanes_mask = np.zeros((n_proposals, n_horizon), dtype=np.bool_)
        batch_multiple_lanes_mask = (
                                            corners_in_polygon[:, :, drivable_lane_idcs].sum(axis=-1) > 0
                                    ).sum(axis=-1) > 1

        batch_not_single_lanes_mask = np.zeros((n_proposals, n_horizon), dtype=np.bool_)
        batch_not_single_lanes_mask = np.all(
            corners_in_polygon[:, :, drivable_lane_idcs].sum(axis=-1) != 4, axis=-1
        )

        multiple_lanes_mask = np.logical_and(batch_multiple_lanes_mask, batch_not_single_lanes_mask)
        self._ego_areas[multiple_lanes_mask, EgoAreaIndex.MULTIPLE_LANES] = True

        # in_nondrivable_area: if at least one corner is not within any drivable polygon
        batch_nondrivable_area_mask = np.zeros((n_proposals, n_horizon), dtype=np.bool_)
        batch_nondrivable_area_mask = (
                                              corners_in_polygon[:, :, drivable_area_idcs].sum(axis=-2) > 0
                                      ).sum(axis=-1) < 4
        self._ego_areas[batch_nondrivable_area_mask, EgoAreaIndex.NON_DRIVABLE_AREA] = True

        # in_oncoming_traffic: if center not in any drivable polygon that is on-route
        batch_oncoming_traffic_mask = np.zeros((n_proposals, n_horizon), dtype=np.bool_)
        batch_oncoming_traffic_mask = (
                center_in_polygon[..., drivable_on_route_idcs].sum(axis=-1) == 0
        )
        self._ego_areas[batch_oncoming_traffic_mask, EgoAreaIndex.ONCOMING_TRAFFIC] = True

    def _calculate_no_at_fault_collision(self) -> None:
        """
        Re-implementation of nuPlan's at-fault collision metric.
        """
        no_collision_scores = np.ones(self._num_proposals, dtype=np.float64)

        proposal_collided_track_ids = {
            proposal_idx: copy.deepcopy(self._observation.collided_track_ids)
            for proposal_idx in range(self._num_proposals)
        }

        for time_idx in range(self.proposal_sampling.num_poses + 1):
            ego_polygons = self._ego_polygons[:, time_idx]
            intersecting = self._observation[time_idx].query(ego_polygons, predicate="intersects")

            if len(intersecting) == 0:
                continue

            for proposal_idx, geometry_idx in zip(intersecting[0], intersecting[1]):
                token = self._observation[time_idx].tokens[geometry_idx]
                if (self._observation.red_light_token in token) or (
                        token in proposal_collided_track_ids[proposal_idx]
                ):
                    continue
                ego_in_multiple_lanes_or_nondrivable_area = (
                        self._ego_areas[proposal_idx, time_idx, EgoAreaIndex.MULTIPLE_LANES]
                        or self._ego_areas[proposal_idx, time_idx, EgoAreaIndex.NON_DRIVABLE_AREA]
                )

                tracked_object = self._observation.unique_objects[token]

                # classify collision
                collision_type: CollisionType = get_collision_type(
                    self._states[proposal_idx, time_idx],
                    self._ego_polygons[proposal_idx, time_idx],
                    tracked_object,
                    self._observation[time_idx][token],
                )
                collisions_at_stopped_track_or_active_front: bool = collision_type in [
                    CollisionType.ACTIVE_FRONT_COLLISION,
                    CollisionType.STOPPED_TRACK_COLLISION,
                ]
                collision_at_lateral: bool = (
                        collision_type == CollisionType.ACTIVE_LATERAL_COLLISION
                )

                # 1. at fault collision
                if collisions_at_stopped_track_or_active_front or (
                        ego_in_multiple_lanes_or_nondrivable_area and collision_at_lateral
                ):
                    no_at_fault_collision_score = (
                        0.0 if tracked_object.tracked_object_type in AGENT_TYPES else 0.5
                    )
                    no_collision_scores[proposal_idx] = np.minimum(
                        no_collision_scores[proposal_idx], no_at_fault_collision_score
                    )
                    self._collision_time_idcs[proposal_idx] = min(
                        time_idx, self._collision_time_idcs[proposal_idx]
                    )

                else:  # 2. no at fault collision
                    proposal_collided_track_ids[proposal_idx].append(token)

        self._multi_metrics[MultiMetricIndex.NO_COLLISION] = no_collision_scores

    def _calculate_drivable_area_compliance(self) -> None:
        """
        Re-implementation of nuPlan's drivable area compliance metric
        """
        drivable_area_compliance_scores = np.ones(self._num_proposals, dtype=np.float64)
        off_road_mask = self._ego_areas[:, :, EgoAreaIndex.NON_DRIVABLE_AREA].any(axis=-1)
        drivable_area_compliance_scores[off_road_mask] = 0.0
        self._multi_metrics[MultiMetricIndex.DRIVABLE_AREA] = drivable_area_compliance_scores

    def _calculate_driving_direction_compliance(self) -> None:
        """
        Re-implementation of nuPlan's driving direction compliance metric
        """
        self._multi_metrics[MultiMetricIndex.DRIVING_DIRECTION] = calc_driving_direction_compliance(
            self.metric_cache.others['map_mpi'],
            self._ego_coords,
            BBCoordsIndex,
            self._num_proposals,
            self._config,
            self.proposal_sampling,
        )

    def _calculate_progress(self) -> None:
        """
        Re-implementation of nuPlan's progress metric (non-normalized).
        Calculates progress along the centerline.
        """

        # calculate raw progress in meter
        progress_in_meter = np.zeros(self._num_proposals, dtype=np.float64)
        for proposal_idx in range(self._num_proposals):
            start_point = Point(*self._ego_coords[proposal_idx, 0, BBCoordsIndex.CENTER])
            end_point = Point(*self._ego_coords[proposal_idx, -1, BBCoordsIndex.CENTER])
            progress = self._centerline.project([start_point, end_point])
            progress_in_meter[proposal_idx] = progress[1] - progress[0]
            # print(start_point.x, start_point.y, end_point.x, end_point.y)

        self._progress_raw = np.clip(progress_in_meter, a_min=0, a_max=None)

    def _calculate_ttc(self):
        """
        Re-implementation of nuPlan's time-to-collision metric.
        """

        ttc_scores = np.ones(self._num_proposals, dtype=np.float64)
        temp_collided_track_ids = {
            proposal_idx: copy.deepcopy(self._observation.collided_track_ids)
            for proposal_idx in range(self._num_proposals)
        }

        # calculate TTC for 1s in the future with less temporal resolution.
        future_time_idcs = np.arange(0, 10, 3)
        n_future_steps = len(future_time_idcs)

        # create polygons for each ego position and 1s future projection
        coords_exterior = self._ego_coords.copy()
        coords_exterior[:, :, BBCoordsIndex.CENTER, :] = coords_exterior[
                                                         :, :, BBCoordsIndex.FRONT_LEFT, :
                                                         ]
        coords_exterior_time_steps = np.repeat(coords_exterior[:, :, None], n_future_steps, axis=2)

        speeds = np.hypot(
            self._states[..., StateIndex.VELOCITY_X],
            self._states[..., StateIndex.VELOCITY_Y],
        )

        dxy_per_s = np.stack(
            [
                np.cos(self._states[..., StateIndex.HEADING]) * speeds,
                np.sin(self._states[..., StateIndex.HEADING]) * speeds,
            ],
            axis=-1,
        )

        for idx, future_time_idx in enumerate(future_time_idcs):
            delta_t = float(future_time_idx) * self.proposal_sampling.interval_length
            coords_exterior_time_steps[:, :, idx] = (
                    coords_exterior_time_steps[:, :, idx] + dxy_per_s[:, :, None] * delta_t
            )

        polygons = creation.polygons(coords_exterior_time_steps)

        # check collision for each proposal and projection
        for time_idx in range(self.proposal_sampling.num_poses + 1):
            for step_idx, future_time_idx in enumerate(future_time_idcs):
                current_time_idx = time_idx + future_time_idx
                polygons_at_time_step = polygons[:, time_idx, step_idx]
                intersecting = self._observation[current_time_idx].query(
                    polygons_at_time_step, predicate="intersects"
                )

                if len(intersecting) == 0:
                    continue

                for proposal_idx, geometry_idx in zip(intersecting[0], intersecting[1]):
                    token = self._observation[current_time_idx].tokens[geometry_idx]
                    if (
                            (self._observation.red_light_token in token)
                            or (token in temp_collided_track_ids[proposal_idx])
                            or (speeds[proposal_idx, time_idx] < self._config.stopped_speed_threshold)
                    ):
                        continue

                    ego_in_multiple_lanes_or_nondrivable_area = (
                            self._ego_areas[proposal_idx, time_idx, EgoAreaIndex.MULTIPLE_LANES]
                            or self._ego_areas[proposal_idx, time_idx, EgoAreaIndex.NON_DRIVABLE_AREA]
                    )
                    ego_rear_axle: StateSE2 = StateSE2(
                        *self._states[proposal_idx, time_idx, StateIndex.STATE_SE2]
                    )

                    centroid = self._observation[current_time_idx][token].centroid
                    track_heading = self._observation.unique_objects[token].box.center.heading
                    track_state = StateSE2(centroid.x, centroid.y, track_heading)
                    # TODO: fix ego_area for intersection
                    if is_agent_ahead(ego_rear_axle, track_state) or (
                            (
                                    ego_in_multiple_lanes_or_nondrivable_area
                                    or self._drivable_area_map.is_in_layer(
                                ego_rear_axle.point, layer=SemanticMapLayer.INTERSECTION
                            )
                            )
                            and not is_agent_behind(ego_rear_axle, track_state)
                    ):
                        ttc_scores[proposal_idx] = np.minimum(ttc_scores[proposal_idx], 0.0)
                        self._ttc_time_idcs[proposal_idx] = min(
                            time_idx, self._ttc_time_idcs[proposal_idx]
                        )
                    else:
                        temp_collided_track_ids[proposal_idx].append(token)

        self._weighted_metrics[WeightedMetricIndex.TTC] = ttc_scores

    def _calculate_is_comfortable(self) -> None:
        """
        Re-implementation of nuPlan's comfortability metric.
        """
        from navsim.agents.expansion.submetrics.metric_extended_comfort import calc_extened_comfort
        self._weighted_metrics[WeightedMetricIndex.COMFORTABLE] = calc_extened_comfort(self._states, self._pre_states,
                                                                                       self.proposal_sampling.num_poses,
                                                                                       self.proposal_sampling.interval_length,
                                                                                       self._num_proposals)
        # time_point_s: npt.NDArray[np.float64] = (
        #         np.arange(0, self.proposal_sampling.num_poses + 1).astype(np.float64)
        #         * self.proposal_sampling.interval_length
        # )
        # is_comfortable = ego_is_comfortable(self._states, time_point_s)
        # self._weighted_metrics[WeightedMetricIndex.COMFORTABLE] = np.all(is_comfortable, axis=-1)


# split=trainval
# scene_filter=navtrain
#
# split=test
# scene_filter=navtest
# python navsim/planning/script/run_metric_caching.py \
# split=${split} \
# scene_filter=${scene_filter} \
# worker=ray_distributed_no_torch \
# +cache.for_lctgen=false \
# worker.threads_per_node=128 \
# cache.cache_path=/zhenxinl_nuplan/navsim_workspace/exp/${scene_filter}_consistency_expansion_metric_cache \
# cache.force_feature_computation=true \
# --config-name default_metric_caching
