from tqdm import tqdm
import traceback
import pickle
import hydra
from hydra.utils import instantiate
from omegaconf import DictConfig
import os

from pathlib import Path
from typing import Dict
import logging

from navsim.agents.abstract_agent import AbstractAgent
from navsim.common.dataclasses import Trajectory, SceneFilter
from navsim.common.dataloader import SceneLoader
from nuplan.planning.simulation.trajectory.trajectory_sampling import TrajectorySampling
import numpy as np

logger = logging.getLogger(__name__)

CONFIG_PATH = "../config/pdm_scoring"
CONFIG_NAME = "default_run_create_submission_pickle_navtest"

# args: pkl_vov, pkl_davit, pkl_moe, ...

@hydra.main(config_path=CONFIG_PATH, config_name=CONFIG_NAME)
def main(cfg: DictConfig) -> None:
    data_path = Path(cfg.navsim_log_path)
    sensor_blobs_path = Path(cfg.sensor_blobs_path)
    save_path = '/result'
    scene_filter = instantiate(cfg.scene_filter)

    pkl_path_vov = f'{os.getenv("NAVSIM_EXP_ROOT")}/ensemble_navtest/vov_trainval_512x2048_epoch17.pkl'
    pkl_path_moe = f'{os.getenv("NAVSIM_EXP_ROOT")}/ensemble_navtest/da+eva+vov_trainval_512x2048_epoch12.pkl'
    pkl_path_davit = f'{os.getenv("NAVSIM_EXP_ROOT")}/ensemble_navtest/davit_trainval_256x1024_epoch16.pkl'

    all_predictions = {
        'vov': pickle.load(open(pkl_path_vov, 'rb')),
        'davit': pickle.load(open(pkl_path_davit, 'rb')),
        'moe': pickle.load(open(pkl_path_moe, 'rb'))
    }

    # token -> argmax index
    weights = {
        'vov': {
            'imi': 0.02,
            'noc': 0.7,
            'da': 0.1,
            'ttc': 5.0,
            'progress': 5.0,
            'comfort': 2.0,
            'tpc': 8.0
        },
        'moe': {
            'imi': 0.03,
            'noc': 0.001,
            'da': 0.024,
            'ttc': 5.0,
            'progress': 5.0,
            'comfort': 2.0,
            'tpc': 7.0
        },
        'davit': {
            'imi': 0.02,
            'noc': 0.6,
            'da': 0.5,
            'ttc': 5.0,
            'progress': 5.0,
            'comfort': 2.0,
            'tpc': 3.0
        },
    }
    prop_vov = 0.5
    prop_davit = 0.4
    prop_moe = 0.1
    ensembled_scores = {}
    for token, v_vov in all_predictions['vov'].items():
        v_davit = all_predictions['davit'][token]
        v_moe = all_predictions['moe'][token]
        chosen_traj_index = prop_vov * (
                weights['vov']['imi'] * v_vov['imi'] +
                weights['vov']['noc'] * v_vov['noc'] +
                weights['vov']['da'] * v_vov['da'] +
                weights['vov']['tpc'] * np.log(
                        weights['vov']['ttc'] * np.exp(v_vov['ttc']) +
                        weights['vov']['progress'] * np.exp(v_vov['progress']) +
                        weights['vov']['comfort'] * np.exp(v_vov['comfort'])
                )) + prop_davit * (
                    weights['davit']['imi'] * v_davit['imi'] +
                    weights['davit']['noc'] * v_davit['noc'] +
                    weights['davit']['da'] * v_davit['da'] +
                    weights['davit']['tpc'] * np.log(
                        weights['davit']['ttc'] * np.exp(v_davit['ttc']) +
                        weights['davit']['progress'] * np.exp(v_davit['progress']) +
                        weights['davit']['comfort'] * np.exp(v_davit['comfort'])
                )) + prop_moe * (
                    weights['moe']['imi'] * v_moe['imi'] +
                    weights['moe']['noc'] * v_moe['noc'] +
                    weights['moe']['da'] * v_moe['da'] +
                    weights['moe']['tpc'] * np.log(
                        weights['moe']['ttc'] * np.exp(v_moe['ttc']) +
                        weights['moe']['progress'] * np.exp(v_moe['progress']) +
                        weights['moe']['comfort'] * np.exp(v_moe['comfort'])
                ))
        ensembled_scores[token] = chosen_traj_index.argmax(0)


    output = ensemble_subscores_to_pickle(
        scene_filter=scene_filter,
        data_path=data_path,
        sensor_blobs_path=sensor_blobs_path,
        vocab=np.load(f'{os.getenv("NAVSIM_DEVKIT_ROOT")}/traj_final/test_8192_kmeans.npy'),
        ensembled_scores=ensembled_scores
    )

    submission = {
        "team_name": "<a href=https://github.com/NVlabs/Hydra-MDP>Hydra-MDP</a>",
        "authors": "Zhenxin Li, Kailin Li, Shihao Wang, Shiyi Lan, Zhiding Yu, Zhiqi Li, Yishen Ji, Ziyue Zhu, Jan Kautz, Jose M. Alvarez",
        "email": "23210240025@m.fudan.edu.cn",
        "institution": "NVIDIA, Nankai University, Nanjing University, Fudan University, East China Normal University, Beijing Institute of Technology",
        "country / region": "United States, China",
        "predictions": [output],
    }
    
    # pickle and save dict
    filename = os.path.join(save_path, "submission.pkl")
    with open(filename, 'wb') as file:
        pickle.dump(submission, file)
    logger.info(f"Your submission filed was saved to {filename}")

def ensemble_subscores_to_pickle(
    scene_filter: SceneFilter,
    data_path: Path,
    sensor_blobs_path: Path,
    vocab,
    ensembled_scores
) -> Dict[str, Trajectory]:
    """
    Function to create the output file for evaluation of an agent on the testserver
    :param agent: Agent object
    :param data_path: pathlib path to navsim logs
    :param sensor_blobs_path: pathlib path to sensor blobs
    :param save_path: pathlib path to folder where scores are stored as .csv
    """    
    logger.info("Building Agent Input Loader")
    input_loader = SceneLoader(
        data_path=data_path,
        scene_filter=scene_filter,
        sensor_blobs_path=sensor_blobs_path,
    )
    output: Dict[str, Trajectory] = {}
    for token in tqdm(input_loader, desc="Running evaluation"):
        traj = vocab[ensembled_scores[token]]
        trajectory = Trajectory(traj,
                          TrajectorySampling(time_horizon=4, interval_length=0.1))
        output.update({token: trajectory})

    return output

if __name__ == "__main__":
    main()
