import logging
import lzma
import os
import pickle
import traceback
import uuid
from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Union, Tuple

import hydra
import pandas as pd
import pytorch_lightning as pl
import torch.distributed as dist
from hydra.utils import instantiate
from nuplan.planning.script.builders.logging_builder import build_logger
from nuplan.planning.utils.multithreading.worker_utils import worker_map
from omegaconf import DictConfig
from torch.utils.data import DataLoader

from navsim.agents.abstract_agent import AbstractAgent
from navsim.common.dataloader import MetricCacheLoader
from navsim.common.dataloader import SceneLoader, SceneFilter
from navsim.evaluate.pdm_score import pdm_score
from navsim.planning.metric_caching.metric_cache import MetricCache
from navsim.planning.script.builders.worker_pool_builder import build_worker
from navsim.planning.simulation.planner.pdm_planner.scoring.pdm_scorer import PDMScorer
from navsim.planning.simulation.planner.pdm_planner.simulation.pdm_simulator import (
    PDMSimulator
)
from navsim.planning.training.agent_lightning_module import AgentLightningModule
from navsim.planning.training.dataset import Dataset


"""
ckpt -> pkl + valid score

"""

logger = logging.getLogger(__name__)

CONFIG_PATH = "../config/pdm_scoring"
CONFIG_NAME = "default_run_create_submission_pickle_ddp"

# args: subscore_path


@hydra.main(config_path=CONFIG_PATH, config_name=CONFIG_NAME)
def main(cfg: DictConfig) -> None:
    build_logger(cfg)
    # gpu inference
    agent: AbstractAgent = instantiate(cfg.agent)
    agent.initialize()
    # Extract scenes based on scene-loader to know which tokens to distribute across workers
    # TODO: infer the tokens per log from metadata, to not have to load metric cache and scenes here
    scene_filter = instantiate(cfg.scene_filter)
    scene_loader = SceneLoader(
        sensor_blobs_path=Path(cfg.sensor_blobs_path),
        data_path=Path(cfg.navsim_log_path),
        scene_filter=scene_filter,
        sensor_config=agent.get_sensor_config(),
    )

    private_data = Dataset(
        scene_loader=scene_loader,
        feature_builders=agent.get_feature_builders(),
        target_builders=[],
        cache_path=None,
        force_cache_computation=False,
        agent_input_only=True,
        append_token_to_batch=True
    )

    val_dataloader = DataLoader(private_data, **cfg.dataloader.params, shuffle=False)
    logger.info("Num private samples: %d", len(private_data))

    trainer = pl.Trainer(**cfg.trainer.params, callbacks=agent.get_training_callbacks())

    logger.info("Starting Training")
    predictions = trainer.predict(
        AgentLightningModule(
            agent=agent,
        ),
        val_dataloader,
        return_predictions=True
    )
    dist.barrier()
    all_predictions = [None for _ in range(dist.get_world_size())]

    if dist.is_initialized():
        dist.all_gather_object(all_predictions, predictions)
    else:
        all_predictions.append(predictions)

    # todo put predictions in data_points
    if dist.get_rank() == 0:
        merged_predictions = {}
        for proc_prediction in all_predictions:
            for d in proc_prediction:
                merged_predictions.update(d)
        pickle.dump(merged_predictions, open(f'{cfg.subscore_path}', 'wb'))


if __name__ == "__main__":
    main()
